@@ -2,29 +2,30 @@ import struct
 import base64
 import logging
 from io import BytesIO, SEEK_CUR, SEEK_END
-from typing import Union, Tuple
+from typing import Union, Tuple, TypeVar, Type, Dict
 
 from luckydonaldUtils.encoding import to_unicode
 from luckydonaldUtils.exceptions import assert_type_or_raise
 
 logger = logging.getLogger(__name__)
+CLASS = TypeVar('CLASS')
 
 
-def base64url_decode(string):
+def base64url_decode(string: str) -> bytes:
     # add missing padding # http://stackoverflow.com/a/9807138
     return base64.urlsafe_b64decode(string + '='*(4 - len(string)%4))
 # end def
 
 
-def base64url_encode(string):
+def base64url_encode(string: bytes) -> str:
     return to_unicode(base64.urlsafe_b64encode(string)).rstrip("=")
 # end def
 
 
-def rle_decode(binary) -> bytearray:
+def rle_decode(binary: bytes) -> bytearray:
     """
     Returns the byte array of the given string.
-    :param string:
+    :param binary: Input data.
     :return: The array of ints.
     :rtype: bytearray
     """
@@ -52,7 +53,7 @@ def rle_decode(binary) -> bytearray:
 # end def
 
 
-def rle_encode(binary):
+def rle_encode(binary: bytes) -> bytearray:
     # https://github.com/danog/MadelineProto/blob/1485d3879296a997d47f54469b0dd518b9230b06/src/danog/MadelineProto/TL/Files.php#L85
     new = bytearray()
     count = 0
@@ -68,11 +69,15 @@ def rle_encode(binary):
             new.append(cur)
         # end if
     # end for
+    if count > 0:
+        new.append(0)
+        new.append(count)
+    # end if
     return new
 # end def
 
 
-def posmod(a, b):
+def pos_mod(a: int, b: int) -> int:
     """
     Positive modulo
     Works just like the % (modulus) operator, only returns always a postive number.
@@ -92,11 +97,11 @@ def pack_tl_string(string: Union[str, bytes]) -> bytes:
     concat = b''
     if length <= 253:
         concat += chr(length).encode()
-        fill = posmod(-length - 1, 4)
+        fill = pos_mod(-length - 1, 4)
     else:
         concat += chr(254).encode()
         concat += struct.pack('<L', length)[0:3]
-        fill = posmod(-length, 4)
+        fill = pos_mod(-length, 4)
     # end if
     concat += string
     concat += b'\x00' * fill
@@ -108,10 +113,12 @@ def pack_tl_string(string: Union[str, bytes]) -> bytes:
 
 def unpack_tl_string(buffer: BytesIO, as_string: bool = False) -> Union[str, bytes]:
     """
-    https://github.com/danog/tg-file-decoder/blob/afcdb9a4a7239e36e8ab3b9a02db72eaa95db66e/src/type.php#L395
-    :param data:
-    :return:
+    Unpack a tl_string.
+    :param buffer: Input buffer. Needs support for `.read(1)` and `.seek(SEEK_CUR)`.
+    :param as_string: if we should return it as utf-8 decoded `str` instead of `bytes`.
+    :return: The unpacked string.
     """
+    # https://github.com/danog/tg-file-decoder/blob/afcdb9a4a7239e36e8ab3b9a02db72eaa95db66e/src/type.php#L395
     length: int = ord(buffer.read(1))
 
     if length > 254:
@@ -121,9 +128,9 @@ def unpack_tl_string(buffer: BytesIO, as_string: bool = False) -> Union[str, byt
     if length == 254:
         # untested
         length = int(struct.unpack('<L', (buffer.read(3) + b'\x00'))[0])
-        fill = posmod(-1 * length, 4)
+        fill = pos_mod(-1 * length, 4)
     else:
-        fill = posmod(-(length + 1), 4)
+        fill = pos_mod(-(length + 1), 4)
     # end if
     string = buffer.read(length)
     buffer.seek(fill, SEEK_CUR)
@@ -135,6 +142,12 @@ def unpack_tl_string(buffer: BytesIO, as_string: bool = False) -> Union[str, byt
 
 
 def unpack_null_terminated_string(buffer: BytesIO, as_string: bool = False) -> Union[str, bytes]:
+    """
+    Unpack a null terminated (\0) string.
+    :param buffer: Input buffer. Needs support for `.read(1)`.
+    :param as_string: if we should return it as utf-8 decoded `str` instead of `bytes`.
+    :return: The unpacked string.
+    """
     char = buffer.read(1)
     new_buff = b''
     while char != b'\00':
@@ -148,15 +161,17 @@ def unpack_null_terminated_string(buffer: BytesIO, as_string: bool = False) -> U
 # end def
 
 
-def pack_null_terminated_string(string: Union[str, bytes], as_string: bool = False) -> str:
+def pack_null_terminated_string(string: Union[str, bytes]) -> bytes:
+    """
+    Packs a string and appends the null terminator \0 to it.
+    :param string: The string to encode.
+    :param as_string: if we should return it as utf-8 decoded `str` instead of `bytes`.
+    :return: The unpacked string.
+    """
     if isinstance(string, str):
         string: bytes = string.encode('utf-8')
     # end if
     string = string + b'\00'
-
-    if as_string:
-        string = string.decode('utf-8')
-    # end if
     return string
 # end def
 
@@ -227,7 +242,7 @@ class FileId(object):
     # end def __init__
 
     @property
-    def owner_id(self):
+    def owner_id(self) -> int:
         if not (self.version in (2, 4) and self.type_id == DocumentFileId.TYPE_STICKER):
             return None
         # end if
@@ -244,7 +259,7 @@ class FileId(object):
     # end def
 
     @classmethod
-    def from_file_id(cls, file_id, decoded=None):
+    def from_file_id(cls, file_id, decoded: Union[None, bytes] = None) -> Union['PhotoFileId', 'DocumentFileId', 'WebLocationFileId']:
         """
 
         :param file_id:
@@ -341,16 +356,41 @@ class FileId(object):
         return file_id_obj
     # end def
 
-    def recalculate(self):
+    def recalculate(self) -> str:
         """ Recalculates the file_id """
-        file_id = self.to_file_id()
-        self.file_id = file_id
-        return file_id
+        return self.to_file_id(recalculate=True)
     # end def
 
-    def to_file_id(self, version=None, sub_version=None):
+    def to_file_id(self, *, recalculate: bool = False, version: Union[int, None] = None, sub_version: Union[int, None] = None) -> str:
         """
-        Subclasses do calculation here.
+        Get a file_id.
+        If we already have a cached one present (and `force_recalculate` is False), we will return that one instead.
+        :param recalculate: if we should force a new calculation of it. Optional, default is False (unless it wasn't cached before, or the version differs).
+        :param version:
+        :param sub_version:
+        :return:
+        """
+        if version and self.version != version:
+            # version is different, we need to calculate it.
+            return self.calculate_file_id(version=version, sub_version=sub_version)
+        # end if
+        if sub_version and self.sub_version != sub_version:
+            # version is different, we need to calculate it.
+            return self.calculate_file_id(version=version, sub_version=sub_version)
+        # end if
+
+        if not recalculate and not self.file_id:
+            self.file_id = self.calculate_file_id()
+        # end if
+        return self.file_id
+    # end def
+
+    def calculate_file_id(self, *, version: Union[int, None] = None, sub_version: Union[int, None] = None) -> str:
+        """
+        Calculates a new file id from our fields.
+
+        :param version: supply a different version
+        :param sub_version: supply a different version
         :return:
         """
         if not version:
@@ -359,10 +399,8 @@ class FileId(object):
         if not sub_version:
             sub_version = self.sub_version
         # end if
+        assert (version, sub_version) in self.SUPPORTED_VERSIONS
 
-        if self.file_id:
-            return self.file_id
-        # end if
         type_id = self.type_id
         if self.file_reference:
             type_id |= self.TYPE_ID_FILE_REFERENCE_FLAG
@@ -380,9 +418,7 @@ class FileId(object):
             assert isinstance(self, WebLocationFileId)
             file_id += pack_tl_string(self.url)
             file_id += struct.pack('<iq', self.access_hash)
-            return file_id
-            # buffer.seek(0)
-            # return base64url_encode(rle_encode(buffer.read()))
+            return base64url_encode(rle_encode(file_id))
         # end if
 
         file_id += struct.pack('<qq', self.id, self.access_hash)
@@ -416,9 +452,10 @@ class FileId(object):
             file_id += chr(sub_version).encode()
         # end if
         file_id += chr(version).encode()
+        return base64url_encode(rle_encode(file_id))
     # end def
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "FileId(file_id={file_id!r}, type_id={type_id!r}, type_generic={type_generic!r}, type_detailed={type_detailed!r}, dc_id={dc_id!r}, id={id!r}, access_hash={access_hash!r}, version={version!r}, owner_id={owner_id!r})".format(
             file_id=self.file_id, type_id=self.type_id, type_generic=self.type_generic, type_detailed=self.type_detailed,
             dc_id=self.dc_id, id=self.id, access_hash=self.access_hash, version=self.version,
@@ -447,7 +484,7 @@ class FileId(object):
     TYPE_SIZE = 17
     TYPE_NONE = 18
 
-    SUPPORTED_VERSIONS = (
+    SUPPORTED_VERSIONS: Tuple[Tuple[int, int], ...] = (
         (2, 0),
         (4, 22),
         (4, 27),
@@ -490,7 +527,7 @@ class DocumentFileId(FileId):
             type_detailed: str,
             dc_id: int, id: int, access_hash: int,
             version=2, sub_version=0,
-        ):
+    ):
         """
         :param file_id: Telegram file_id
         :type  file_id: str
@@ -533,21 +570,16 @@ class DocumentFileId(FileId):
         )
     # end def __init__
 
-    TYPES = {
+    TYPES: Dict[int, str] = {
         FileId.TYPE_VOICE: "voice", FileId.TYPE_VIDEO: "video", FileId.TYPE_DOCUMENT: "document",
         FileId.TYPE_STICKER: "sticker", FileId.TYPE_AUDIO: "song", FileId.TYPE_VIDEO_NOTE: "video note",
         FileId.TYPE_ANIMATION: "animation",
     }
     """ A human readable string of the type """
 
-    def swap_type_sticker(self):
+    def swap_type_sticker(self) -> str:
         """
-        This swaps out the document types "document" <-> "sticker"
-
-        :param data: Can be a dict as obtained by `take_apart_file_id(file_id)`.
-                     Otherwise a file_id is assumed and said function `take_apart_file_id` is called.
-        :type  data: dict or bytearray or bytes
-
+        This swaps out the document types "document" <-> "sticker". Modifies the underlying data.
         :return: new file id
         :rtype: str
         """
@@ -556,20 +588,20 @@ class DocumentFileId(FileId):
         return self.recalculate()
     # end def
 
-    def change_type(self, type_id):
+    def change_type(self, type_id: int):
         """
-        Changes the type of the document.
+        Changes the type of the document to the given type
 
         :param type_id:
         :return:
         """
+        self.type_detailed = DocumentFileId.TYPES[self.type_id]  # this raises KeyError if it isn't a valid type.
         self.type_id = type_id
-        self.type_detailed = DocumentFileId.TYPES[self.type_id]
         return self.recalculate()
     # end def
 
     @classmethod
-    def from_file_id(cls, file_id, decoded=None):
+    def from_file_id(cls: Type[CLASS], file_id, decoded: Union[None, bytes] = None) -> Union[FileId, CLASS]:
         """
         :param file_id:
         :param decoded:
@@ -578,43 +610,7 @@ class DocumentFileId(FileId):
         return FileId.from_file_id(file_id=file_id, decoded=decoded)
     # end def
 
-    def to_file_id(self, version=None, sub_version=None):
-        return super().to_file_id(version=version, sub_version=sub_version)
-        assert self.type_id in DocumentFileId.TYPES
-        if version is None:
-            version = self.version
-            sub_version = self.sub_version
-        # end if
-        if version is None:
-            version, sub_version = self.MAX_VERSION
-        # end if
-        if version == 2:
-            binary = struct.pack(
-                "<iiqqb",
-                # type, dc_id, id,
-                self.type_id, self.dc_id, self.id if self.id else 0,
-                # access_hash
-                self.access_hash if self.access_hash else 0,
-                # version
-                2
-            )
-        elif version == 4 and sub_version == 22:
-            binary = struct.pack(
-                "<iiqqbb",
-                # type, dc_id, id,
-                self.type_id, self.dc_id, self.id if self.id else 0,
-                # access_hash,
-                self.access_hash if self.access_hash else 0,
-                # twentytwo, version
-                22, 4
-            )
-        else:
-            raise ValueError(f'Unknown version flag: ({version},{sub_version})')
-        # end if
-        return base64url_encode(rle_encode(binary))
-    # end def
-
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "DocumentFileId(file_id={file_id!r}, type_id={type_id!r}, type_generic={type_generic!r}, type_detailed={type_detailed!r}, dc_id={dc_id!r}, id={id!r}, access_hash={access_hash!r}, version={version!r}, owner_id={owner_id!r})".format(
             file_id=self.file_id, type_id=self.type_id, type_generic=self.type_generic, type_detailed=self.type_detailed,
             dc_id=self.dc_id, id=self.id, access_hash=self.access_hash, owner_id=self.owner_id, version=self.version,
@@ -646,11 +642,11 @@ class WebLocationFileId(object):  # TODO make proper (FileId) subclass:
         self.access_hash = access_hash
     # end def __init__
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"{self.__class__.__name__}()"
     # end def __repr__
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self.__repr__()
     # end def __str__
 # end class WebLocationFileId
@@ -664,11 +660,11 @@ class PhotoFileId(FileId):
             self.location_local_id: int = location_local_id
         # end def __init__
 
-        def __repr__(self):
+        def __repr__(self) -> str:
             return f"{self.__class__.__name__}(type_id={self.type_id!r}, volume_id={self.volume_id!r}, location_local_id={self.location_local_id!r})"
         # end def __repr__
 
-        def __str__(self):
+        def __str__(self) -> str:
             return self.__repr__()
         # end def __str__
     # end class PhotosizeSource
@@ -679,7 +675,7 @@ class PhotoFileId(FileId):
             super().__init__(PhotoFileId.PHOTOSIZE_SOURCE_LEGACY, volume_id=volume_id, location_local_id=location_local_id)
         # end def __init__
 
-        def __repr__(self):
+        def __repr__(self) -> str:
             return f"{self.__class__.__name__}(type_id={self.type_id!r}, volume_id={self.volume_id!r}, location_local_id={self.location_local_id!r}, secret={self.secret})"
         # end def __repr__
     # end class PhotosizeSourceLegacy
@@ -691,7 +687,7 @@ class PhotoFileId(FileId):
             super().__init__(PhotoFileId.PHOTOSIZE_SOURCE_THUMBNAIL, volume_id=volume_id, location_local_id=location_local_id)
         # end def __init__
 
-        def __repr__(self):
+        def __repr__(self) -> str:
             return (
                 f"{self.__class__.__name__}("
                 f"type_id={self.type_id!r}, volume_id={self.volume_id!r}, location_local_id={self.location_local_id!r}, "
@@ -708,7 +704,7 @@ class PhotoFileId(FileId):
             super().__init__(type_id, volume_id=volume_id, location_local_id=location_local_id)
         # end def __init__
 
-        def __repr__(self):
+        def __repr__(self) -> str:
             return (
                 f"{self.__class__.__name__}("
                 f"type_id={self.type_id!r}, volume_id={self.volume_id!r}, location_local_id={self.location_local_id!r}, "
@@ -737,7 +733,7 @@ class PhotoFileId(FileId):
             super().__init__(PhotoFileId.PHOTOSIZE_SOURCE_STICKERSET_THUMBNAIL, volume_id=volume_id, location_local_id=location_local_id)
         # end def __init__
 
-        def __repr__(self):
+        def __repr__(self) -> str:
             return (
                 f"{self.__class__.__name__}("
                 f"type_id={self.type_id!r}, volume_id={self.volume_id!r}, location_local_id={self.location_local_id!r}, "
@@ -786,7 +782,7 @@ class PhotoFileId(FileId):
         self.photosize = photosize
     # end def __init__
 
-    TYPES = {FileId.TYPE_THUMBNAIL: "thumbnail", FileId.TYPE_PROFILE_PHOTO: "profile picture", FileId.TYPE_PHOTO: "photo"}
+    TYPES: Dict[int, str] = {FileId.TYPE_THUMBNAIL: "thumbnail", FileId.TYPE_PROFILE_PHOTO: "profile picture", FileId.TYPE_PHOTO: "photo"}
     """ A human readable string of the type """
 
     PHOTOSIZE_SOURCE_LEGACY: int = 0
@@ -805,50 +801,11 @@ class PhotoFileId(FileId):
     """ Used for document and photo thumbnails """
 
     @classmethod
-    def from_file_id(cls, file_id, decoded=None):
+    def from_file_id(cls: Type[CLASS], file_id, decoded: Union[None, bytes] = None) -> Union[FileId, CLASS]:
         return FileId.from_file_id(file_id=file_id, decoded=decoded)
     # end def
 
-    def to_file_id(self, version=None, sub_version=None):
-        return super().to_file_id(version=version, sub_version=sub_version)
-        assert self.type_id in PhotoFileId.TYPES
-        if version == 2:
-            binary = struct.pack(
-                '<iiqqqqib',
-                # type, dc_id, id,
-                self.type_id, self.dc_id, self.id if self.id else 0,
-                # access_hash,
-                self.access_hash if self.access_hash else 0,
-                # location_volume_id, location_secret,
-                self.location.volume_id, self.location.secret,
-                # location_local_id,
-                self.location.local_id,
-                # version
-                2
-            )
-        elif version == 4:
-            binary = struct.pack(
-                '<iiqqqqiibb',
-                # type, dc_id, id,
-                self.type_id, self.dc_id, self.id if self.id else 0,
-                # access_hash,
-                self.access_hash if self.access_hash else 0,
-                # location_volume_id, location_secret,
-                self.location.volume_id, self.location.secret,
-                # something,
-                self.something,
-                #  location_local_id,
-                self.location.local_id,
-                # twentytwo, version
-                22, 4
-            )
-        else:
-            raise ValueError(f'Unknown version to use: {version}')
-        # end if
-        return base64url_encode(rle_encode(binary))
-    # end def
-
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"PhotoFileId(" \
                f"file_id={self.file_id!r}, type_id={self.type_id!r}, type_generic={self.type_generic!r}, " \
                f"type_detailed={self.type_detailed!r}, dc_id={self.dc_id!r}, id={self.id!r}, " \
@@ -857,4 +814,3 @@ class PhotoFileId(FileId):
                f")"
     # end def __str__
 # end class PhotoFileId
-# end def
